//
// Rendering functions
// Most functions support alpha blending
//

#include "renderlib.h"
#include "render.h"
#include "surface.h"
#include "font.h"
#include "utils.h"


// Alpha blending function
//
__inline uint32_t renderPixel(const uint32_t Dest, const uint32_t Src, const uint32_t blendOp)
{

	switch (blendOp)
	{
		// Solid color, set alpha to full
		case BLEND_SOLID:
			return (Src & 0x00FFFFFF) | 0xFF000000;

		// Fixed 50% alpha blend
		case BLEND_ALPHA50:
			if (ALPHA(Src) == 0xFF)
				return Dest;
			else
				return (((Dest & 0x00FEFEFE) >> 1) + ((Src & 0x00FEFEFE) >> 1)) | (0xFF << 24);

		// Full alpha blending
		case BLEND_ALPHA:
			if (ALPHA(Src) == 0)
				return Dest;
			if (ALPHA(Src) == 0xFF)
				return Src;
			else
			{
				uint8_t  A = ALPHA(Src);
				uint32_t RB = (((Src & 0x00FF00FF) * A) + ((Dest & 0x00FF00FF) * (0xFF - A))) & 0xFF00FF00;
				uint32_t G =  (((Src & 0x0000FF00) * A) + ((Dest & 0x0000FF00) * (0xFF - A))) & 0x00FF0000;

				return (0xFF << 24) | (RB | G) >> 8;
			}

		// Simple alpha, only return pixel if source has alpha set to full
		case BLEND_ALPHA_SIMPLE:
			if (ALPHA(Src) == 0xFF)
				return Src;
			else
				return Dest;

		// Invert destination
		case BLEND_INVERT:
			if (ALPHA(Src) == 0xFF)
				return Dest;

			return Dest ^ Src;

	}

	return 0;

}


// Render a line
//
void EXPORT renderLine(const Surface* destSurf, int32_t X1, int32_t Y1, const int32_t X2, const int32_t Y2, const uint32_t Color, const uint32_t blendOp)
{

	register int32_t deltaX = abs(X2 - X1) << 1;
	register int32_t deltaY = abs(Y2 - Y1) << 1;

	// if X1 == X2 or Y1 == Y2, then it does not matter what we set here
	register uint8_t iX = X2 > X1 ? 1 : -1;
	register uint8_t iY = Y2 > Y1 ? 1 : -1;


	if (!destSurf)
		return;

	if (X1 >= 0 && X1 < destSurf->Width && Y1 >= 0 && Y1 < destSurf->Height)
		*(destSurf->ptrRows[Y1] + X1) = renderPixel(*(destSurf->ptrRows[Y1] + X1), Color, blendOp);

	if (deltaX >= deltaY)
	{
		// error may go below zero
		register int32_t Error = deltaY - (deltaX >> 1);

		while (X1 != X2)
		{
			if (Error >= 0)
			{
				if (Error || (iX > 0))
				{
					Y1 += iY;
					Error -= deltaX;
				}
			}

			X1 += iX;
			Error += deltaY;

			if ((uint32_t)X1 <= destSurf->Width && (uint32_t)Y1 < destSurf->Height)
				*(destSurf->ptrRows[Y1] + X1) = renderPixel(*(destSurf->ptrRows[Y1] + X1), Color, blendOp);
		}
	}
	else
	{
		// error may go below zero
		register int32_t Error = deltaX - (deltaY >> 1);

		while (Y1 != Y2)
		{
			if (Error >= 0)
			{
				if (Error || (iY > 0))
				{
					X1 += iX;
					Error -= deltaY;
				}
			}

			Y1 += iY;
			Error += deltaX;

			if ((uint32_t)X1 <= destSurf->Width && (uint32_t)Y1 < destSurf->Height)
				*(destSurf->ptrRows[Y1] + X1) = renderPixel(*(destSurf->ptrRows[Y1] + X1), Color, blendOp);
		}
	}


}


// Render text to a surface from a font surface
//
void EXPORT renderText(const Surface* destSurf, const Font* srcFont, const int32_t x, const int32_t y, const uint8_t* Text, uint32_t Color, const uint32_t blendOp)
{

	register int32_t   rX, cX, cY;

	int32_t   dX, dY;
	uint32_t* Dest;
	uint32_t  cChar = 0;


	if (!destSurf || !srcFont || !Text)
		return;

	if (y + srcFont->charHeight < 0 || y > destSurf->Height)
		return;

	// Strip alpha
	Color &= 0x00FFFFFF;

	do
	{
		// Rendering everything above space characters
		if (*Text > 32)
		{
			// Calculate X coordinate on font surface
			rX = (*Text - 33) * srcFont->charWidth;

			for (cY = 0; cY < srcFont->charHeight; cY++)
			{
				for(cX = 0; cX < srcFont->charWidth; cX++)
				{
					// Calculate coordinates on destination surface
					dX = x + cX + (cChar * srcFont->charWidth);
					dY = (y + cY);

					if ((uint32_t)dX < destSurf->Width && (uint32_t)dY < destSurf->Height)
					{
						Dest = destSurf->ptrRows[dY] + dX;
						*Dest = renderPixel(*Dest, Color | *(srcFont->Surface->ptrRows[cY] + cX + rX), blendOp);
					}
				}
			}
		}

		cChar++;
		Text++;
	} while (*Text != 0);

}


// Render a box (no fill)
//
void EXPORT renderBox(const Surface* destSurf, const int32_t x, const int32_t y, const int32_t Width, const int32_t Height, const uint32_t Color, const uint32_t blendOp)
{

	if (!destSurf)
		return;

	renderLine(destSurf, x,             y,              x + Width - 2, y,              Color, blendOp);
	renderLine(destSurf, x,             y + 1,          x,             y + Height - 2, Color, blendOp);
	renderLine(destSurf, x + Width - 1, y,              x + Width - 1, y + Height - 2, Color, blendOp);
	renderLine(destSurf, x,             y + Height - 1, x + Width - 1, y + Height - 1, Color, blendOp);

}


// Initialize the renderer
//
void EXPORT renderInit()
{

}


// Blit one surface onto another, no scaling, no alpha blending
//
void EXPORT renderBlitFast(const Surface* destSurf, const Surface* srcSurf, int32_t X, int32_t Y)
{

	uint32_t* Dest;
	uint32_t* Src;

	int32_t  cY, Len;
	int32_t  rW, rH, xO, yO;


	if (!destSurf || !srcSurf)
		return;

	rW = srcSurf->Width;
	rH = srcSurf->Height;
	xO = yO = 0;

	// Bounds checks
	if (X < 0)
	{
		rW = rW + X;
		if (rW <= 0)
			return;
		xO = -X;
		X = 0;
	}

	if (Y < 0)
	{
		rH = rH + Y;
		if (rH <= 0)
			return;
		yO = -Y;
		Y = 0;
	}

	if (X + rW > destSurf->Width)
	{
		rW = destSurf->Width - X;
		if (rW <= 0)
			return;
	}

	if (Y + rH > destSurf->Height)
	{
		rH = destSurf->Height - Y;
		if (rH <= 0)
			return;
	}


	// Blit row by row
	Dest = destSurf->ptrRows[Y] + X;
	Src = srcSurf->ptrRows[yO] + xO;
	Len = rW << 2;
	for (cY = 0; cY < rH; cY++)
	{
		memcpy(Dest, Src, Len);
		Dest += destSurf->Width;
		Src += srcSurf->Width;
	}

}


// Render a filled box
//
void EXPORT renderBoxFill(const Surface* destSurf, int32_t X, int32_t Y, int32_t Width, int32_t Height, const uint32_t Color, const uint32_t blendOp)
{

	register int32_t cX, cY;

	uint32_t *Dest;
	uint32_t *destRow;


	if (!destSurf)
		return;

	// Bounds check
	if (X < 0)
	{
		Width = Width + X;
		if (Width <= 0)
			return;
		X = 0;
	}

	if (Y < 0)
	{
		Height = Height + Y;
		if (Height <= 0)
			return;
		Y = 0;
	}

	if (X + Width > destSurf->Width)
	{
		Width = destSurf->Width - X;
		if (Width <= 0)
			return;
	}

	if (Y + Height > destSurf->Height)
	{
		Height = destSurf->Height - Y;
		if (Height <= 0)
			return;
	}

	// Render
	destRow = destSurf->ptrRows[Y] + X;
	for (cY = 0; cY < Height; cY++)
	{
		Dest = destRow;
		for (cX = 0; cX < Width; cX++)
		{
			*Dest = renderPixel(*Dest, Color, blendOp);
			Dest++;
		}

		destRow += destSurf->Width;
	}

}


// Blit a surface onto another. Only alpha blending, no scaling.
//
void EXPORT renderBlitAlphaFast(const Surface* destSurf, const Surface *srcSurf, int32_t X, int32_t Y, const uint32_t blendOp)
{

	register int32_t cX, cY;

	int32_t    rW, rH, xO, yO;
	uint32_t*  Dest;
	uint32_t*  Src;


	if (!destSurf || !srcSurf)
		return;

	rW = srcSurf->Width;
	rH = srcSurf->Height;
	xO = yO = 0;

	// Bounds checks
	if (X < 0)
	{
		rW = rW + X;
		if (rW <= 0)
			return;
		xO = -X;
		X = 0;
	}

	if (Y < 0)
	{
		rH = rH + Y;
		if (rH <= 0)
			return;
		yO = -Y;
		Y = 0;
	}

	if (X + rW > destSurf->Width)
	{
		rW = destSurf->Width - X;
		if (rW <= 0)
			return;
	}

	if (Y + rH > destSurf->Height)
	{
		rH = destSurf->Height - Y;
		if (rH <= 0)
			return;
	}


	// Blit
	for (cY = 0; cY < rH; cY++)
	{
		Dest = destSurf->ptrRows[Y + cY] + X;
		Src = srcSurf->ptrRows[yO + cY] + xO;

		for (cX = 0; cX < rW; cX++)
		{
			*Dest = renderPixel(*Dest, *Src, blendOp);
			Src++;
			Dest++;
		}
	}

}


// Blit a surface onto another. Does scaling and alpha blending.
//
void EXPORT renderBlit(const Surface* destSurf, const Surface* srcSurf, const int32_t X, const int32_t Y, const int32_t Width, const int32_t Height, const uint32_t blendOp)
{

	register int32_t  cX, cY;
	int32_t   cU, cV, xRatio, yRatio;
	uint32_t* Dest;


	if (!srcSurf || !destSurf || Width < 1 || Height < 1)
		return;

	xRatio  = (srcSurf->Width << 16) / Width;
	yRatio  = (srcSurf->Height << 16) / Height;

	cV = 0;

	for(cY = Y; cY < Y + Height; cY++)
	{
		if ((uint32_t)cY < destSurf->Height)
		{
			cU = 0;

			for(cX = X; cX < X + Width; cX++)
			{
				if ((uint32_t)cX < destSurf->Width)
				{
					Dest = destSurf->ptrRows[cY] + cX;
					*Dest = renderPixel(*Dest, srcSurf->ptrRows[cV >> 16][cU >> 16], blendOp);
				}

				cU += xRatio;
			}
		}

		cV += yRatio;
	}

}