//
// Surfaces
// A surface is a 32-bit BGRA image.
//
#include "renderlib.h"
#include "surface.h"
#include "png.h"
#include "utils.h"


// Flip surface vertically
//
Surface* EXPORT surfaceFlipY(const Surface* srcSurf)
{

	uint32_t Y;
	Surface* destSurf;

	uint32_t* Dest;
	uint32_t* Src;
	uint32_t Len;


	if (!srcSurf)
		return NULL;

	destSurf = surfaceCreate(srcSurf->Width, srcSurf->Height);
	if (!destSurf)
		return NULL;

	Dest = destSurf->ptrData;
	Src = srcSurf->ptrRows[srcSurf->Height - 1];
	Len = srcSurf->Width << 2;
	for (Y = 0; Y < srcSurf->Height; Y++)
	{
		memcpy(Dest, Src, Len);
		Dest += destSurf->Width;
		Src -= srcSurf->Width;
	}

	return destSurf;

}


// Return a surface's properties
//
uint32_t EXPORT surfaceGetWidth(const Surface* Surf)
{
	return Surf->Width;
}

uint32_t EXPORT surfaceGetHeight(const Surface* Surf)
{
	return Surf->Height;
}


// Write a surface to a PNG file
//
uint32_t EXPORT surfaceWriteToPNG(const Surface* Surf, const char* fileName, const uint8_t compressLevel)
{

	PNGContext* png;
	FILE* fp;
	uint32_t* imgData;
	uint32_t X;
	uint32_t Y;
	uint32_t Src;


	if (!Surf)
		return 0;

	if (compressLevel < 1 || compressLevel > 9)
		return 0;

	png = pngCreate();
	png->Width = Surf->Width;
	png->Height = Surf->Height;
	png->colorType = PNG_COLOR_TYPE_RGB_ALPHA;
	png->bitDepth = 8;

	// Convert image data to RGBA
	imgData = malloc(png->Width * png->Height * sizeof(uint32_t));
	for (Y = 0; Y < Surf->Height; Y++)
	{
		for (X = 0; X < Surf->Width; X++)
		{
			Src = *(Surf->ptrRows[Y] + X);
			Src = BGRA(RED(Src), GREEN(Src), BLUE(Src), ALPHA(Src));
			imgData[X + Y * png->Width] = Src;
		}
	}

	pngSave(png, imgData, NULL, 0, compressLevel);

	fp = fopen(fileName, "wb");
	if (!fp)
	{
		free(imgData);
		pngClose(png);
		return 0;
	}
	if (fwrite(png->Data, 1, png->dataSize, fp) == 0)
	{
		free(imgData);
		fclose(fp);
		pngClose(png);
		return 0;
	}

	free(imgData);
	fclose(fp);
	pngClose(png);

	return 1;

}


// Load a surface from a PNG file
//
Surface* EXPORT surfaceReadFromPNG(const char* fileName)
{

	PNGContext* png;
	Surface* Surf;
	FILE* fp;
	void* Data;
	uint8_t* imgData;
	uint32_t dataLen;

	uint32_t X;
	uint32_t Y;


	// Read file
	fp = fopen(fileName, "rb");
	fseek(fp, 0, SEEK_END);
	dataLen = ftell(fp);
	fseek(fp, 0, SEEK_SET);

	Data = malloc(dataLen);
	if (fread(Data, 1, dataLen, fp) == 0)
	{
		free(Data);
		fclose(fp);
		return NULL;
	}

	// Read PNG
	png = pngOpen(Data, dataLen);
	free(Data);
	fclose(fp);
	if (!png)
		return NULL;

	// Only 8 bits per channel is supported
	if (png->bitDepth != 8)
	{
		pngClose(png);
		return NULL;
	}

	// Create new surface
	Surf = surfaceCreate(png->Width, png->Height);
	if (!Surf)
	{
		pngClose(png);
		return NULL;
	}

	// Greyscale becomes an alpha channel
	if (png->bitDepth == 8 && png->colorType == PNG_COLOR_TYPE_GRAY)
	{
		imgData = malloc(png->Width * png->Height);
		pngRead(png, imgData, NULL);

		for(Y = 0; Y < Surf->Height; Y++)
			for(X = 0; X < Surf->Width; X++)
				*(Surf->ptrRows[Y] + X) = *(imgData + X + Y * Surf->Width) << 24;

		free(imgData);
	}
	else
	{
		pngClose(png);
		surfaceDestroy(Surf);
		return NULL;
	}

	pngClose(png);

	return Surf;

}


// Allocate a new, empty surface
//
Surface* EXPORT surfaceCreate(const uint32_t Width, const uint32_t Height)
{

	uint32_t	Row;
	Surface*	Surf;


	Surf = malloc(sizeof(Surface));
	Surf->Width   = Width;
	Surf->Height  = Height;
	Surf->dataLen = Surf->Width * Surf->Height * sizeof(uint32_t);
	Surf->ptrData = calloc(1, Surf->dataLen);
	Surf->ptrRows = malloc(Surf->Height * sizeof(uint32_t));

	// Assign row pointers
	for (Row = 0; Row < Surf->Height; Row++)
		Surf->ptrRows[Row] = (Surf->ptrData + Row * Surf->Width);

	return Surf;

}


// Destroy a surface
//
void EXPORT surfaceDestroy(Surface* Surf)
{

	if (!Surf)
		return;

	free(Surf->ptrRows);
	free(Surf->ptrData);
	free(Surf);

	Surf = NULL;

}


// Copy a surface to a regular DIB, doubling each pixel
//
void EXPORT surfaceCopyToBitmapDouble(const Surface* srcSurf, uint32_t* ptrBitmapBits, const uint32_t destWidth, const uint32_t destHeight)
{

	register uint32_t   cU, cV;

	uint32_t   cX, cY;
	uint32_t*  destRow;
	uint32_t*  Dest;


	destRow = ptrBitmapBits + (destHeight * destWidth) - destWidth;
	cV = 0;

	cY = destHeight - 1;
	while(cY--)
	{
		cU = 0;
		Dest = destRow;

		cX = destWidth - 1;
		while(cX--)
		{
			*Dest++ = srcSurf->ptrRows[cV >> 16][cU >> 16];
			cU += 0x8000;
		}

		cV += 0x8000;
		destRow -= destWidth;
	}

}


// Copy a surface to a regular DIB
//
void EXPORT surfaceCopyToBitmap(const Surface *srcSurf, uint32_t *ptrBitmapBits)
{

	register uint32_t  *Src, *Dest;

	int32_t   y, Len;


	Len = srcSurf->Width << 2;
	Src = srcSurf->ptrRows[srcSurf->Height - 1];
	Dest = ptrBitmapBits;

	for (y = 0; y < srcSurf->Height; y++)
	{
		memcpy(Dest, Src, Len);
		Src -= srcSurf->Width;
		Dest += srcSurf->Width;
	}

}


// Get a pixel from a surface
//
uint32_t EXPORT surfaceGetPixel(const Surface* srcSurf, const int32_t X, const int32_t Y)
{

	if ((uint32_t)X >= srcSurf->Width || (uint32_t)Y >= srcSurf->Height || !srcSurf)
		return 0;

	return *(srcSurf->ptrRows[Y] + X);

}


// Set a pixel to this surface
//
void EXPORT surfaceSetPixel(const Surface* destSurf, const int32_t X, const int32_t Y, const uint32_t Color)
{

	if ((uint32_t)X >= destSurf->Width || (uint32_t)Y >= destSurf->Height || !destSurf)
		return;

	*(destSurf->ptrRows[Y] + X) = Color;

}


// Extract a portion of a surface onto another
//
void EXPORT surfaceExtract(const Surface* srcSurf, const Surface* destSurf, const int32_t X, const int32_t Y)
{

	int32_t		Row, Len;
	uint32_t*	Src;
	uint32_t*	Dest;


	Len = destSurf->Width << 2;
	Src = srcSurf->ptrRows[Y] + X;
	Dest = destSurf->ptrData;

	for (Row = 0; Row < destSurf->Height; Row++)
	{
		memcpy(Dest, Src, Len);
		Src += srcSurf->Width;
		Dest += destSurf->Width;
	}

}


// Fill this surface with a single color
//
void EXPORT surfaceFill(const Surface* destSurf, const uint32_t Color)
{

	int32_t	Pix;


	if (!destSurf) return;

	for(Pix = 0; Pix < destSurf->Width * destSurf->Height; Pix++)
		*(destSurf->ptrData + Pix) = Color;

}


// Clear a surface
//
void EXPORT surfaceClear(const Surface* Surf)
{

	memset(Surf->ptrData, 0, Surf->dataLen);

}